/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "icmp/icmppacketfactory.h"

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "icmp/icmpchecksum.h"
#include "icmp/icmpdata.h"
#include "icmp/icmptype.h"
#include "icmp/icmpv4header.h"
#include "icmp/icmpv4packet.h"
#include "icmp/icmpv6header.h"
#include "icmp/icmpv6packet.h"

using namespace std;
using boost::asio::ip::icmp;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// IcmpPacketFactory
//-----------------------------------------------------------------------------

/**
 * @brief Creates an ICMP packet from the input stream @c std::istream.
 *
 * @param protocol The packet's network layer protocol, IPv4 or IPv6.
 * @param is The input stream.
 *
 * @return An ICMP Packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmp_packet(
        const icmp::socket::protocol_type &protocol,
        istream &is
)
{
    IcmpPacketItem icmp_packet;

    if ( icmp::v4() == protocol )
    {
        icmp_packet.reset( new Icmpv4Packet() );
    }
    else if ( icmp::v6() == protocol )
    {
        icmp_packet.reset( new Icmpv6Packet() );
    }
    else
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << "Unknown protocol, expect ICMP v4 or v6!" << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
        return icmp_packet;
    }

    IcmpPacket::ReadReturnCode return_code = icmp_packet->read( is );
    if ( return_code != IcmpPacket::ReadReturnCode_OK )
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << IcmpPacket::return_code_to_string(return_code) << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
           // --> icmp_pinger will not try to continue working with this packet
    }
    else if ( !is.good() )
    {
        GlobalLogger.warning() << "ICMP packet creation failed: "
            << "Stream not good at end of creation!" << endl;
        icmp_packet.reset();    // --> (!icmp_packet) is true
    }

    return icmp_packet;
}

/**
 * @brief Creates an ICMP Echo Request packet.
 *
 * @param protocol The packet's network layer protocol, IPv4 or IPv6.
 * @param identifier The packet's identifier number to aid in matching Echo
 * Replies to this Echo Request. May be zero.
 * @param sequence_number The packet's sequence number to aid in matching Echo
 * Replies to this Echo Request. May be zero.
 *
 * @return An ICMP Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmp_packet_echo_request(
        const icmp::socket::protocol_type &protocol,
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    BOOST_ASSERT( (icmp::v4() == protocol) || (icmp::v6() == protocol) );

    IcmpPacketItem icmp_packet;

    if ( icmp::v4() == protocol )
    {
        icmp_packet = create_icmpv4_packet_echo_request( identifier, sequence_number );
    }
    else if ( icmp::v6() == protocol )
    {
        icmp_packet = create_icmpv6_packet_echo_request( identifier, sequence_number );
    }
    else
    {
        BOOST_ASSERT( !"Invalid ICMP Packet Type." );                                                                                               //lint !e506
    }

    return icmp_packet;
}

/**
 * @brief Creates an ICMPv4 Echo Request packet.
 *
 * @param identifier The packet's identifier number.
 * @param sequence_number The packet's sequence number.
 *
 * @return An ICMPv4 Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmpv4_packet_echo_request(
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    const IcmpData icmp_data( "ping-message" );

    Icmpv4Type type = Icmpv4Type_EchoRequest;
    uint8_t code = 0;
    IcmpChecksum calculator( icmp_data.begin(), icmp_data.end() );
    uint16_t checksum = calculator.compute(
            type, code, identifier, sequence_number
    );
    Icmpv4Header icmp_header(
            type, code, checksum, identifier, sequence_number
    );
    IcmpPacketItem icmp_packet( new Icmpv4Packet( icmp_header, icmp_data ) );

    return icmp_packet;
}

/**
 * @brief Creates an ICMPv6 Echo Request packet.
 *
 * @param identifier The packet's identifier number.
 * @param sequence_number The packet's sequence number.
 *
 * @return An ICMPv6 Echo Request packet object.
 */
IcmpPacketItem IcmpPacketFactory::create_icmpv6_packet_echo_request(
        const uint16_t identifier,
        const uint16_t sequence_number
)
{
    const IcmpData icmp_data( "ping-message" );

    Icmpv6Type type = Icmpv6Type_EchoRequest;
    uint8_t code = 0;
    IcmpChecksum calculator( icmp_data.begin(), icmp_data.end() );
    uint16_t checksum = calculator.compute(
            type, code, identifier, sequence_number
    );
    Icmpv6Header icmp_header(
            type, code, checksum, identifier, sequence_number
    );
    IcmpPacketItem icmp_packet( new Icmpv6Packet( icmp_header, icmp_data ) );

    return icmp_packet;
}


